import os
import pathlib
from enum import Enum


class BuildType(Enum):
    CATKIN_MAKE = 1
    CATKIN_TOOLS = 2
    COLCON = 3


def _get_parent_dirs(cur_dir):
    """Iterate over all parent directories (including the starting directory)."""
    folder = cur_dir.resolve()
    while folder:
        yield folder
        if folder.parent == folder:
            return
        else:
            folder = folder.parent


def get_workspace_root(cur_dir=pathlib.Path('.')):
    """
    Return the workspace root and the type of workspace.

    If no build tool has been run yet, return the highest directory with a src subfolder
    If catkin_make or catkin build has been run, return the directory with the proper metadata.
    If colcon, there's no metadata, so we return the highest directory with a src subfolder AND a build subfolder.
    """
    prefix_path = os.environ.get('COLCON_PREFIX_PATH')
    if prefix_path:
        first = prefix_path.split(':')[0]
        install_dir = pathlib.Path(first)
        return BuildType.COLCON, install_dir.parent
    highest_candidate = None
    for folder in _get_parent_dirs(cur_dir):
        if (folder / '.catkin_workspace').exists():
            return BuildType.CATKIN_MAKE, folder
        elif (folder / '.catkin_tools').exists():
            return BuildType.CATKIN_TOOLS, folder
        elif (folder / 'src').exists():
            if (folder / 'build').exists():
                # If we're here, it is not CATKIN
                return BuildType.COLCON, folder
            highest_candidate = folder

    return None, highest_candidate


def get_ros_version(fail_quietly=False):
    """Return ROS Version and Distro."""
    env = os.environ
    version = env.get('ROS_VERSION')
    distro = env.get('ROS_DISTRO')
    if version and distro:
        return int(version), distro

    if fail_quietly:
        return version, distro

    raise RuntimeError('Unable to determine ROS distro. Please source workspace.')



import pathlib

from .workspace import get_ros_version
from .packages import get_package_names

INTERFACE_TYPES = ['msg', 'srv', 'action']


class ROSInterface:
    def __init__(self, package, interface_type, name):
        self.package = package
        self.type = interface_type
        self.name = name

    @staticmethod
    def from_string(s, interface_type=None):
        pieces = s.split('/')
        # If only two pieces, assume the interface_type is missing
        # i.e. convert geometry_msgs/Point to geometry_msgs/msg/Point
        if len(pieces) == 2:
            return ROSInterface(pieces[0], interface_type, pieces[1])
        # If three pieces, assume the interface type is specified
        elif len(pieces) == 3:
            return ROSInterface(*pieces)
        else:
            raise RuntimeError(f'Cannot parse interface for "{s}"')

    def to_string(self, two_part=True):
        if two_part:
            return f'{self.package}/{self.name}'
        else:
            return f'{self.package}/{self.type}/{self.name}'

    def __lt__(self, other):
        if self.package != other.package:
            return self.package < other.package

        i0 = INTERFACE_TYPES.index(self.type)
        i1 = INTERFACE_TYPES.index(other.type)
        if i0 != i1:
            return i0 < i1

        return self.name < other.name

    def __repr__(self):
        return self.to_string()


def list_interfaces(ros_version=None, interface_types=None):
    """yields ROSInterface instances for all messages/services/actions in the environment."""
    if ros_version is None:
        ros_version, _ = get_ros_version()
    if interface_types is None:
        interface_types = INTERFACE_TYPES

    if ros_version == 1:
        import rospkg
        from rosmsg import _list_types, _get_package_paths

        rospack = rospkg.RosPack()
        for pkg in get_package_names(ros_version):
            package_paths = _get_package_paths(pkg, rospack)
            for package_path_s in package_paths:
                package_path = pathlib.Path(package_path_s)
                for interface_type in interface_types:
                    interface_path = package_path / interface_type
                    if not interface_path.exists():
                        continue
                    for i_name in _list_types(str(interface_path),
                                              interface_type,
                                              '.' + interface_type):
                        yield ROSInterface(pkg, interface_type, i_name)
    else:
        from rosidl_runtime_py import get_interfaces
        for pkg, pkg_interfaces in get_interfaces().items():
            for pkg_interface in pkg_interfaces:
                interface_type, i_name = pkg_interface.split('/')
                if interface_type not in interface_types:
                    continue
                yield ROSInterface(pkg, interface_type, i_name)



import pathlib
import re

from .workspace import get_ros_version, get_workspace_root, _get_parent_dirs

PACKAGE_NAME = re.compile(r'<name>(.*)</name>')


def get_package_name(folder):
    """If this folder is the root of a package, return the name of the package."""
    filename = folder / 'package.xml'
    if filename.exists():
        s = open(filename).read()
        m = PACKAGE_NAME.search(s)
        if m:
            return m.group(1)
        else:
            return folder.stem


def get_package_name_from_path(initial_path=pathlib.Path('.')):
    """Return the name of the package that the initial path is in (including subfolders)."""
    for folder in _get_parent_dirs(initial_path):
        pkg_name = get_package_name(folder)
        if pkg_name:
            return pkg_name


def get_package_names(ros_version=None):
    """returns a set of strings of all the current package names

    including the binary packages and the workspace packages"""
    if ros_version is None:
        ros_version, _ = get_ros_version()

    if ros_version == 1:
        import rospkg
        rospack = rospkg.RosPack()
        return set(rospack.list())
    else:
        from ament_index_python import get_packages_with_prefixes
        return set(get_packages_with_prefixes().keys())


class _FakeNamespace:
    """Used to fake argparser values"""

    def __init__(self, **kwargs):
        self.__dict__.update(kwargs)

    def __getattr__(self, name):
        return self.__dict__.get(name, [])


def is_binary_ros(path):
    return str(path).startswith('/opt/ros/')


_CACHED_DESCRIPTORS = {}


def find_package(package_name, ros_version=None, workspace_root=None):
    if ros_version is None:
        ros_version, _ = get_ros_version()

    if ros_version == 1:
        import rospkg
        rospack = rospkg.RosPack()
        return pathlib.Path(rospack.get_path(package_name))
    else:
        from ros2pkg.api import get_prefix_path
        prefix = get_prefix_path(package_name)

        if prefix and is_binary_ros(prefix):
            return pathlib.Path(f'{prefix}/share/{package_name}')
        else:
            from colcon_core.package_selection import get_package_descriptors

            if workspace_root is None:
                _, workspace_root = get_workspace_root()

            if workspace_root not in _CACHED_DESCRIPTORS:
                args = _FakeNamespace(base_paths=[str(workspace_root)], ignore_user_meta=True)
                descriptors = get_package_descriptors(args)
                _CACHED_DESCRIPTORS[workspace_root] = descriptors
            else:
                descriptors = _CACHED_DESCRIPTORS[workspace_root]

            for descriptor in descriptors:
                if descriptor.name == package_name and descriptor.path:
                    return pathlib.Path(descriptor.path)

            # Install-only path
            return prefix



from .workspace import get_ros_version, get_workspace_root, BuildType
from .packages import get_package_name_from_path, get_package_names, find_package, is_binary_ros
from .interfaces import list_interfaces, ROSInterface

__all__ = ['get_ros_version', 'get_workspace_root', 'get_package_name_from_path',
           'BuildType', 'get_package_names', 'find_package', 'is_binary_ros', 'list_interfaces', 'ROSInterface']
